(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import ssr countable vgtac basis lang_common lang_spec lang_mid.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".



(** * Miscellaneous. *)


Lemma three_way_perm A (x y z : list A) :
  Permutation (y ++ x ++ z) (x ++ y ++ z).
Proof.
  rewrite !app_assoc; apply Permutation_app; auto.
  apply Permutation_app_comm; auto.  
Qed.


Ltac comm_to_start X := 
    try (eapply Permutation_trans; [
         by repeat first [
             apply (Permutation_app_comm _ X) 
           |  apply (Permutation_app_head) ] |]);
    first [ apply Permutation_app_head |
            eapply Permutation_trans; [
              by repeat first [
                   apply (three_way_perm X) 
                |  apply (Permutation_app_head) ] |]; comm_to_start X ].


Ltac crush_perm :=
  by repeat first [rewrite <- !app_assoc
                  | rewrite !(cons_app _ (_ :: _))
                  | rewrite !(cons_app _ (_ ++ _))
                  | match goal with
                        |- Permutation _ (?x ++ _) => comm_to_start x
                    end].


Ltac normalise_list :=
  repeat first [ rewrite <- !app_assoc
               | rewrite !(cons_app _ (_ :: _))
               | rewrite !(cons_app _ (_ ++ _))].

Lemma Permutation_app_comm_useful (l1 l1' l2 l2' : list var) :
  Permutation l1 l1' -> Permutation l2 l2' -> Permutation l2' l1' -> Permutation l1 l2.
Proof. ii; eauto using Permutation_sym, Permutation_trans. Qed.

Lemma Permutation_wf_bvar (bv vs vs' G : list var):
  Permutation vs vs' ->
  Permutation (vs ++ bv ++ G) (bv ++ vs' ++ G).
Proof.
  ii. rewrite !app_assoc. apply Permutation_app; eauto. 
  eapply Permutation_app_comm_useful.
  - eapply Permutation_app_comm.
  - eapply Permutation_refl.
  - eapply Permutation_app; eauto using Permutation_sym.
Qed.


(** * Renamings
*)

Record renaming := 
  { rename_termvar : tmap (X:=termvar) (label + termvar)
  ; rename_contvar : tmap (X:=contvar) contvar
  }.

Definition ren_extend_t (r : renaming) x y : renaming :=
  {| rename_termvar := [y//x] >> (rename_termvar r);
     rename_contvar := rename_contvar r  |}.

Definition ren_extend_c (r : renaming) x y : renaming :=
  {| rename_contvar := [y//x] >> (rename_contvar r);
     rename_termvar := rename_termvar r  |}.

Definition ren_purge (r : renaming) : renaming :=
  {| rename_contvar := fun (x : contvar) => x;
     rename_termvar := rename_termvar r |}.

Notation "r '[' y '/' x ']' " := (ren_extend_t r x y) (at level 0, y at level 10, 
                                                        format "r [ y / x ]"
                                                       ).
Notation "r '[c' y '/' x ']' " := (ren_extend_c r x y) (at level 0, y at level 10,
                                                        format "r [c  y / x ]"
                                                       ).

Notation "'|' r '|'" := (ren_purge r) (at level 11, format "| r |").

Definition rename_itvar r (x : itermvar) : itermvar := 
  match x with
    | inl lx => inl lx
    | inr tx => rename_termvar r tx
  end.

Lemma ren_purge_it r x :
  rename_itvar (|r|) x = rename_itvar r x.
Proof. by destruct x. Qed.

Definition itvar2var x : var :=
  inl x.

Lemma itvar2var_tkvar x :
  itvar2var (inr x) = tkvar2var (inl x).
Proof. reflexivity. Qed.

Definition rename_tkvar r (x : termvar + contvar) : var :=
  match x with
    | inl tx => inl (rename_termvar r tx)
    | inr tk => inr (rename_contvar r tk)
  end.

Definition rename_var r (x : var) : var :=
  match x with
    | inl ltx => inl (rename_itvar r ltx)
    | inr kx => inr (rename_contvar r kx)
  end.

Lemma ren_ext_it_neq : forall r x y z
  (NEQ: z <> inr x), rename_itvar r[y/x] z = rename_itvar r z.
Proof. ii; s; destruct z; auto; s; rewrite tmext_neq; eauto. Qed.

Lemma ren_ext_it_eq : forall r x y,
                    rename_itvar r [y/x] (inr x) = y.
Proof. s. i; by rewrite tmext_eq. Qed.
Hint Resolve ren_ext_it_eq.

Lemma ren_ext_eq : forall r x y,
                    rename_var r [y/x] (inl (inr x)) = inl y.
Proof. s. i. by rewrite tmext_eq. Qed.

Lemma ren_ext_c_eq : forall r x y,
                       rename_var r [c y/x] (inr x) = inr y.
Proof. s. i. by rewrite tmext_eq. Qed.

Lemma ren_ext_neq : forall r x y z
  (NEQ: z <> inl (inr x)), rename_var r[y/x] z = rename_var r z.
Proof. ii; s; destruct z as [[z|z]|z]; auto; s. rewrite tmext_neq; ii; clarify. Qed.

Lemma ren_ext_c_neq : forall r x y z,
  rename_contvar r[y/x] z = rename_contvar r z.
Proof. ii; s; destruct z; auto; s; rewrite tmext_neq; eauto. Qed.

Lemma ren_ext_it_neq_c : forall r x y z,
  rename_itvar r[c y/x] z = rename_itvar r z.
Proof. ii; s; destruct z; auto; s; rewrite tmext_neq; eauto. Qed.

Lemma ren_ext_c_neq_c : forall r x y z
  (NEQ: z <> x), rename_contvar r[c y/x] z = rename_contvar r z.
Proof. ii; s; destruct z; auto; s; rewrite tmext_neq; eauto. Qed.
  
Lemma ren_c_ext_neq : forall r x y z (NEQ: z <> inr x), rename_var r[c y/x] z = rename_var r z.
Proof. ii; s; destruct z as [[z|z]|z]; auto; s. rewrite tmext_neq; ii; clarify. Qed.

Lemma ren_ext_c_neq' : forall r x y z (E: z <> x),
  rename_contvar r[c y/x] z = rename_contvar r z.
Proof. ii; s; destruct z; auto; s; rewrite tmext_neq; eauto. Qed.

Lemma ren_var_ext_neq: forall ren x y (E: x <> inr y),
  rename_var (ren)[c y/y] x = rename_var ren x.
Proof.
  i; destruct x; s; rewrite ?tmext_neq; ii; vauto.
Qed.





































  







  





















Definition map_vars : renaming -> exp -> exp :=
  fix f r e :=
    match e with
      | IEunit => IEunit
      | IEnat n => IEnat n
      | IEpair x y => IEpair (rename_itvar r x) (rename_itvar r y)
      | IEinl x => IEinl (rename_itvar r x)
      | IEinr x => IEinr (rename_itvar r x)
      | IEbop bop xn yn => IEbop bop 
                                 (match xn with inl it => inl (rename_itvar r it) | inr n => inr n end) 
                                 (match yn with inl it => inl (rename_itvar r it) | inr n => inr n end) 
      | IEfst x => IEfst (rename_itvar r x)
      | IEsnd x => IEsnd (rename_itvar r x)
      | IEfix bf bx bk ea => IEfix bf bx bk (f (r[inr bf/bf][inr bx/bx][c bk/bk]) ea)
      | IEgen bk ea => IEgen bk (f (r[c bk/bk]) ea)
      | IEcont bx ea => IEcont bx (f r[inr bx/bx] ea)
      | IEnew x y ea => IEnew (rename_itvar r x) y (f r[inr y/y] ea)
      | IEread x y ea => IEread (rename_itvar r x) y (f r[inr y/y] ea)
      | IEwrite x y ea => IEwrite (rename_itvar r x) (rename_itvar r y) (f r ea)
      | IEinput x ea => IEinput x (f r[inr x/x] ea)
      | IEoutput x ea => IEoutput (rename_itvar r x) (f r ea)
      | IEifnz x ea eb => IEifnz (rename_itvar r x) (f r ea) (f r eb)
      | IEcase x y ea eb => IEcase (rename_itvar r x) y (f r[inr y/y] ea) (f r[inr y/y] eb)
      | IEret k x => IEret (rename_contvar r k) (rename_itvar r x)
      | IEapp x y k => IEapp (rename_itvar r x) (rename_itvar r y) (rename_contvar r k)
      | IEinst x k => IEinst (rename_itvar r x) (rename_contvar r k)
      | IElet ea (inl x) eb => IElet (f r ea) (inl x) (f r[inr x/x] eb)
      | IElet ea (inr x) eb => IElet (f r ea) (inr x) (f r[c x/x] eb)
      | IErefeq x y => IErefeq (rename_itvar r x) (rename_itvar r y)
    end.

Definition rename := map_vars.

Lemma rename_bvars r e: bvars (rename r e) = bvars e.
Proof. generalize r; induction e; s; ii; eauto; (try destruct bx); s; rewrite ?IHe ?IHe1 ?IHe2; eauto. Qed.














































Lemma purge_rename_var G r :
  purge (map (rename_var r) G) = map (rename_var r) (purge G).
Proof. induction G; eauto; destruct a as [[]|]; s; eauto; by rewrite IHG. Qed.

Lemma map_rename_var_purge G r :
  map (rename_var (|r|)) (purge G) = map (rename_var r) (purge G).
Proof. induction G; eauto; destruct a as [[]|]; s; eauto; by rewrite IHG. Qed.

Lemma map_rename_var_agree G r1 r2:
  (forall x, In x G -> rename_var r1 x = rename_var r2 x) ->
  map (rename_var r1) G = map (rename_var r2) G.
Proof. induction G; eauto; destruct a as [[a|a]|a]; s; ii; eauto; rewrite IHG; eauto.
  - by injection (H (inl (inr a))); eauto; ii; rewrite H0.
  - by injection (H (inr a)); eauto; ii; rewrite H0.
Qed.
  










































































































































(** ** Identity renamings
*)
Definition ren_id : renaming := {| rename_termvar := inr; rename_contvar := id |}.

Lemma ren_var_id: forall x,
  rename_var ren_id x = x.
Proof.
    by destruct x as [[]|].
Qed.

Lemma ren_itvar_id: forall x,
  rename_itvar ren_id x = x.
Proof.
    by destruct x as [[]|].
Qed.

Lemma ren_contvar_id: forall x,
  rename_contvar ren_id x = x.
Proof.
    by destruct x as [[]|].
Qed.






















                   










































Lemma ren_id_self_t: forall x,
  (ren_id)[inr x/x] = ren_id.
Proof.
  i; unfold ren_id; unfold ren_extend_t; s; f_equal.
  rewrite -(tmext_self x inr); auto.
Qed.

Lemma ren_id_self_c: forall x,
  (ren_id)[c x/x] = ren_id.
Proof.
  i; unfold ren_id; unfold ren_extend_c; s; f_equal.
  rewrite -(tmext_self x id); auto.
Qed.

Lemma rename_id: forall e,
  rename ren_id e = e.
Proof.
  induction e; s; try destruct x; try destruct y; try destruct op
  ; try destruct i; try destruct i0; try destruct xf; try destruct bx
  ; rewrite ?ren_id_self_t ?ren_id_self_c ?IHe ?IHe1 ?IHe2; eauto.
Qed.

Lemma ren_c_cons_t: forall x y (G: list var) ren
    (IN: In (inr x) (inl (inr y) :: G))
    (X: forall (IN: In (inr x) G), In (inr (rename_contvar (ren) x)) G),
  In (inr (rename_contvar (ren)[inr y/y] x)) (inl (inr y) :: G).
Proof.
  i; apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
Qed.

Lemma ren_t_cons_t: forall x y (G: list var) ren
    (IN: In (inl x) (inl (inr y) :: G))
    (X: forall (IN: In (inl x) G), In (inl (rename_itvar (ren) x)) G),
  In (inl (rename_itvar (ren)[inr y/y] x)) (inl (inr y) :: G).
Proof.
  i; apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
  rewrite ren_ext_it_neq; ii; vauto; eauto.
Qed.

Lemma ren_c_cons_c: forall x y (G: list var) ren
    (IN: In (inr x) (inr y :: G))
    (X: forall (IN: In (inr x) G), In (inr (rename_contvar (ren) x)) G),
  In (inr (rename_contvar (ren)[c y/y] x)) (inr y :: G).
Proof.
  i; apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
  rewrite tmext_neq; ii; vauto; eauto.
Qed.

Lemma ren_t_cons_c: forall x y (G: list var) ren
    (IN: In (inl x) (inr y :: G))
    (X: forall (IN: In (inl x) G), In (inl (rename_itvar (ren) x)) G),
  In (inl (rename_itvar (ren)[c y/y] x)) (inr y :: G).
Proof.
  i; apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
Qed.








Inductive renG: renaming -> list var -> list var -> Prop :=
| renG_nil G
  : renG ren_id G G
| renG_cons_t ren G G' x y
    (RG: renG ren G G')
  : renG (ren)[y/x] (inl (inr x) :: G) (inl y :: G')
| renG_cons_c ren G G' x y
    (RG: renG ren G G')
  : renG (ren)[c y/x] (inr x :: G) (inr y :: G')
| renG_cons_purge ren G G' x y
    (RG: renG ren (purge G) (purge G'))
  : renG (ren)[c y/x] (purge G) (purge G')
.
Hint Constructors renG.

Lemma renG_In_c: forall x ren G G'
    (IN: In (inr x) G)
    (RG: renG ren G G'),
  In (inr (rename_contvar ren x)) G'.
Proof.
  i; ginduction RG; i.
  - destruct x; auto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
    s; rewrite tmext_neq; [|by ii; vauto].
    apply IHRG in IN; eauto.
  - exfalso; eauto using purge_inDr.
Qed.

Lemma renG_In_t: forall x ren G G'
    (IN: In (inl x) G)
    (RG: renG ren G G'),
  In (inl (rename_itvar ren x)) G'.
Proof.
  i; ginduction RG; i.
  - destruct x; auto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
    destruct x0; vauto.
    + apply IHRG in IN; s in IN; eauto.
    + s; rewrite tmext_neq; [|by ii; vauto].
      apply IHRG in IN; eauto.
  - apply in_consD in IN; des; clarify; s; tmext_simpl; eauto.
  - eauto.
Qed.

Lemma renG_purge: forall ren G G'
    (RG: renG ren G G'),
  renG ren (purge G) (purge G').
Proof.
  i; ginduction G; i; inv RG; s; eauto; rewrite ?purge_purge; auto.
Qed.


Lemma rename_wfatom_: forall e e' G G' ren
    (WF: wfatom G e)
    (RG: renG ren G G')
    (IH: forall e', size e' < size e -> forall e'' G G' ren
             (WF: wfexp G e') (RG: renG ren G G') (E: rename ren e' = e''),
           wfexp G' e'')
    (E: rename ren e = e'),
  wfatom G' e'.
Proof.
  i; destruct e; try (by vauto)
  ; try (by s in WF; des; s in E; clarify; s; splits
         ; eauto using renG_In_t, renG_In_c)
  ; try (by s in WF; des; s in E; clarify; s; splits
         ; eauto
         ; eapply IH; eauto using renG_purge; s; omega).
  - destruct x, y; s in WF; des; s in E; clarify; s; splits
    ; eauto using renG_In_t, renG_In_c.
Qed.

Lemma rename_wfexp: forall e e' G G' ren
    (WF: wfexp G e)
    (RG: renG ren G G')
    (E: rename ren e = e'),
  wfexp G' e'.
Proof.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i
  ; try (by s in WF; des; s in E; clarify; s; splits
         ; eauto using renG_In_t, renG_In_c
         ; eapply IH; eauto; s; eauto)
  ; try (s in WF; des; s in E; clarify; s; splits
         ; eauto using renG_In_t, renG_In_c
         ; [eapply (IH e1)|eapply (IH e2)]; eauto; s; omega).
  - destruct bx; s in WF; des; s in E; clarify; s; splits; eauto.
    + eapply rename_wfatom_; eauto; i; eapply IH; eauto; s; omega.
    + eapply IH; eauto using ren_t_cons_t, ren_c_cons_t; s; omega.
    + eapply rename_wfatom_; eauto; i; eapply IH; eauto; s; omega.
    + eapply IH; eauto using ren_c_cons_c, ren_t_cons_c; s; omega.
Qed.

Lemma rename_wfatom: forall e e' G G' ren
    (WF: wfatom G e)
    (RG: renG ren G G')
    (E: rename ren e = e'),
  wfatom G' e'.
Proof.
  i; eapply rename_wfatom_; eauto using rename_wfexp.
Qed.
