(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

open Big_int
open Lexer
open Parser
open Printf

open Basics
open Basis
open Compiler
open Datatypes
open Lang_common
open Lang_tgt
open Lang_tgt_interp
open Lang_tgt_syntax
open Lang_src



(** Showing machine syntax. *)


let show_binop = function
  | Plus -> "Plus"
  | Minus -> "Minus"
  | Mult -> "Mult"
  | Equal -> "Equal"
;;

let show_reg = function
  | Coq_r_sp -> "sp"
  | Coq_r_env -> "env"
  | Coq_r_arg -> "arg"
  | Coq_r_clo -> "clo"
  | Coq_r_ret -> "ret"
  | Coq_r_aux -> "aux"
  | Coq_r_i -> "i"
;;

let show_offset = function
  | Coq_o_pos n -> "+" ^ string_of_big_int n
  | Coq_o_neg n -> "-" ^ string_of_big_int n
;;

let show_operand = function
  | Coq_o_imm n -> string_of_big_int n
  | Coq_o_reg r -> show_reg r
  | Coq_o_sind (r, o) -> "{" ^ show_reg r ^ show_offset o ^ "}"
  | Coq_o_hind (r, o) -> "[" ^ show_reg r ^ show_offset o ^ "]"
;;

let show_instr = function
  | Coq_i_jmp o ->
    "jmp " ^ show_operand o
  | Coq_i_jnz (r, o) ->
    "jnz " ^ show_reg r ^ " " ^ show_operand o
  | Coq_i_ld (r, o) ->
    "ld " ^ show_reg r ^ " " ^ show_operand o
  | Coq_i_sto (o, r) ->
    "sto " ^ show_operand o ^ " " ^ show_reg r
  | Coq_i_lpc r ->
    "lpc " ^ show_reg r
  | Coq_i_opr (op, r, o1, o2) ->
    "opr " ^ show_binop op ^ " " ^ show_reg r ^ " "
           ^ show_operand o1 ^ " " ^ show_operand o2
  | Coq_i_inp r ->
    "inp " ^ show_reg r
  | Coq_i_out r ->
    "out " ^ show_reg r
  | Coq_i_alc (r1, r2) ->
    "alc " ^ show_reg r1 ^ " " ^ show_reg r2
;;


(*
let _ =
  let x = Coq_i_jmp (Coq_o_sind (Coq_r_ret, Coq_o_neg 31)) in
  print_endline (string_of_bool (x = decode_instr (encode_instr x)))
*)



(** Main. *)


let big =
  big_int_of_int
;;
  

let laddr =
  big 0
;;


let print_function (lbl, seg) =
  let f i w =
    let s = match decode_instr w with
              | Some x -> show_instr x
              | None -> "INVALID INSTRUCTION" in
    printf "%3s: %s\n" (string_of_big_int i) s;
    succ_big_int i in
  printf "%3s: FUNCTION %s (code starts at %s)\n"
      (string_of_big_int (fst seg))
      (string_of_big_int lbl)
      (string_of_big_int (succ_big_int (fst seg)));
  let _ = List.fold_left f (succ_big_int (fst seg)) (snd seg) in
  print_newline ()
;;

let save_module m filename : unit =
  let cout = open_out_bin filename in
  Marshal.to_channel cout m [Marshal.Closures];
  close_out cout
;;

let load_module filename : Lang_tgt.coq_module =
  let cin = open_in_bin filename in
  let m = (Marshal.from_channel cin : Lang_tgt.coq_module) in
  close_in cin;
  m
;;



let rec execute (cp: computer) n : unit =
  if cphp cp = None
  then printf "\nCrashed at %s (%s steps).\n"
           (string_of_big_int (cppc cp)) (string_of_int n)
  else match step cp with
  | Coq_efc_halted ->
    printf "\nTerminated successfully (%s steps).\n" (string_of_int n)
  | Coq_efc_tau cp' ->
    execute cp' (n+1)
  | Coq_efc_in f ->
    print_string "in: ";
    let m = big_int_of_string (read_line ()) in
    execute (f m) (n+1)
  | Coq_efc_out (m, cp') ->
    printf "out: %s\n" (string_of_big_int m);
    execute cp' (n+1)
;;


let main_link files : unit =
  let m1 = load_module (List.nth files 0) in
  let m2 = load_module (List.nth files 1) in
  save_module (link m1 m2) (List.nth files 2)
;;

let main_awkward files : unit =
  save_module Awkward.mt (List.nth files 0)
;;

let main_compile sel files : unit =
  let cin = open_in (List.nth files 0) in
  let lexbuf = Lexing.from_channel cin in
  let (imports, msrc) = Parser.annotated_modul Lexer.token lexbuf in
  let mtgt = pilsner sel imports msrc in
  save_module mtgt (List.nth files 1)
;;

let main_run files : unit =
  let m = load_module (List.nth files 0) in
  let lsegs = m laddr [] in
  printf "Loaded %n functions:\n\n" (List.length lsegs);
  List.iter print_function lsegs;
  match cload lsegs with
  | Some cp ->
    printf "Executing function %s (code starts at %s).\n\n"
        (string_of_big_int coq_MAIN)
        (string_of_big_int (cppc cp));
    execute cp 0
  | None -> failwith "Loading failed."
;;


type mode = Link | Compile | Run | Awkward
;;

let _ =
  let sel = ref do_all in
  let mode = ref Run in
  let flags =
    [("-c", Arg.Unit (fun _ -> mode := Compile), "")
    ;("-l", Arg.Unit (fun _ -> mode := Link), "")
    ;("-a", Arg.Unit (fun _ -> mode := Awkward), "")
    ;("-no-inline", Arg.Unit (fun _ -> sel := {!sel with do_inline = false}), "")
    ;("-no-contify", Arg.Unit (fun _ -> sel := {!sel with do_contify = false}), "")
    ;("-no-dce", Arg.Unit (fun _ -> sel := {!sel with do_dce = false}), "")
    ;("-no-hoist", Arg.Unit (fun _ -> sel := {!sel with do_hoist = false}), "")
    ;("-no-commute", Arg.Unit (fun _ -> sel := {!sel with do_commute = false}), "")
    ;("-no-dedup", Arg.Unit (fun _ -> sel := {!sel with do_dedup = false}), "")
    ] in
  let files = ref [] in
  Arg.parse flags (fun s -> files := !files @ [s]) "";
  match !mode with
  | Link -> main_link !files
  | Compile -> main_compile !sel !files
  | Run -> main_run !files
  | Awkward -> main_awkward !files
;;
