(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program Relations Omega.
Require Export  paco10  basis vgtac.
Require Import lang_common types lang_spec.
Set Implicit Arguments. 


(** * Queries to a global world. *)

(** Value queries.

    [good_fun], [good_gen], and [good_name] are used only to prove vertical
    compositionality.
*)
Inductive vquery {L: LangSpec} :=
| valid_unit
| valid_nat  (n: nat)
| valid_pair (va: L.(val)) (vb: L.(val))
| valid_inl  (v: L.(val))
| valid_inr  (v: L.(val))
| valid_roll (v: L.(val))
| valid_pack (v: L.(val))
| valid_fun
| valid_gen
| valid_name
| good_fun
| good_gen
.

(** Configuration queries.

    - [app f a k]: the configuration represents an application of
      function [f] to argument [a] in continuation [k]
    - [inst g k]: the configuration represents an instantiation of
      generalization [g] in continuation [k]
    - [ret v k]: the configuration represents the return of value [v]
      to continuation [k]
*)
Inductive cquery {L: LangSpec} :=
| app (f: L.(val)) (a: L.(val)) (k: L.(cont))
| inst (g: L.(val)) (k: L.(cont))
| ret (v: L.(val)) (k: L.(cont))
.


(** * State transition systems. *)

(** Definition.

    - [ext s' s] denotes that [s'] is a future state of [s].
    - [extpub s' s] denotes that [s'] is a _public_ future state of [s].
 *)
Structure transys :=
{ state: Type
; ext: relation state
; extpub: relation state
; ext_trans:
    forall x y (EXT: ext x y) z (EXT': ext y z), ext x z
; extpub_refl:
    forall x, extpub x x 
; extpub_trans:
    forall x y (EXT: extpub x y) z (EXT': extpub y z), extpub x z 
; extpub_in_ext:
    forall x y (EXTP: extpub x y), ext x y
}.
Hint Immediate extpub_refl extpub_in_ext.

Lemma ext_refl (w: transys) :  forall s,
  ext w s s.
Proof.
  i; apply extpub_in_ext, extpub_refl.
Qed.
Hint Immediate ext_refl.

(** Binary product transition system. *)
Program Definition tsprod (tsa tsb: transys) : transys :=
{| state := tsa.(state) * tsb.(state)
 ; ext s' s := tsa.(ext) (fst s') (fst s) /\ 
               tsb.(ext) (snd s') (snd s)
 ; extpub s' s := tsa.(extpub) (fst s') (fst s) /\ 
                  tsb.(extpub) (snd s') (snd s)
|}.
Next Obligation. split; eauto using ext_trans. Qed.
Next Obligation. split; eauto using extpub_trans. Qed.
